"use strict";
/**
 * Tokenizer
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.scanToken = void 0;
const KEYWORDS = [
    'SELECT',
    'INSERT',
    'DELETE',
    'UPDATE',
    'CREATE',
    'DROP',
    'DATABASE',
    'SCHEMA',
    'TABLE',
    'VIEW',
    'TRIGGER',
    'FUNCTION',
    'INDEX',
    'ALTER',
    'TRUNCATE',
    'WITH',
    'AS',
    'MATERIALIZED',
    'BEGIN',
    'DECLARE',
    'CASE',
    'LOOP',
    'IF',
    'REPEAT',
    'WHILE',
    'FOR',
    'PROCEDURE',
    'SHOW',
    'DATABASES',
    'KEYS',
    'TABLES',
    'COLUMNS',
    'STATUS',
    'BINARY',
    'BINLOG',
    'CHARACTER',
    'COLLATION',
    'ENGINE',
    'ENGINES',
    'ERRORS',
    'EVENTS',
    'GRANTS',
    'MASTER',
    'OPEN',
    'PLUGINS',
    'PRIVILEGES',
    'PROCESSLIST',
    'PROFILE',
    'PROFILES',
    'RELAYLOG',
    'REPLICAS',
    'SLAVE',
    'REPLICA',
    'TRIGGERS',
    'VARIABLES',
    'WARNINGS',
];
const INDIVIDUALS = {
    ';': 'semicolon',
};
const ENDTOKENS = {
    '"': '"',
    "'": "'",
    '`': '`',
    '[': ']',
};
function scanToken(state, dialect = 'generic', paramTypes = { positional: true }) {
    const ch = read(state);
    if (isWhitespace(ch)) {
        return scanWhitespace(state);
    }
    if (isCommentInline(ch, state)) {
        return scanCommentInline(state);
    }
    if (isCommentBlock(ch, state)) {
        return scanCommentBlock(state);
    }
    if (isString(ch, dialect) && ch !== null) {
        return scanString(state, ENDTOKENS[ch]);
    }
    if (isParameter(ch, state, paramTypes)) {
        return scanParameter(state, dialect, paramTypes);
    }
    if (isDollarQuotedString(state)) {
        return scanDollarQuotedString(state);
    }
    if (isQuotedIdentifier(ch, dialect) && ch !== null) {
        return scanQuotedIdentifier(state, ENDTOKENS[ch]);
    }
    if (isLetter(ch)) {
        return scanWord(state);
    }
    const individual = scanIndividualCharacter(state);
    if (individual) {
        return individual;
    }
    return skipChar(state);
}
exports.scanToken = scanToken;
function read(state, skip = 0) {
    if (state.position + skip === state.input.length - 1) {
        return null;
    }
    state.position += 1 + skip;
    return state.input[state.position];
}
function unread(state) {
    if (state.position === state.start) {
        return;
    }
    state.position--;
}
function peekBack(state) {
    if (state.position == 0) {
        return null;
    }
    return state.input[state.position - 1];
}
function peek(state) {
    if (state.position >= state.input.length - 1) {
        return null;
    }
    return state.input[state.position + 1];
}
function isKeyword(word) {
    return KEYWORDS.includes(word.toUpperCase());
}
function resolveIndividualTokenType(ch) {
    return INDIVIDUALS[ch];
}
function scanWhitespace(state) {
    let nextChar;
    do {
        nextChar = read(state);
    } while (isWhitespace(nextChar));
    if (nextChar !== null && !isWhitespace(nextChar)) {
        unread(state);
    }
    const value = state.input.slice(state.start, state.position + 1);
    return {
        type: 'whitespace',
        value,
        start: state.start,
        end: state.start + value.length - 1,
    };
}
function scanCommentInline(state) {
    let nextChar;
    do {
        nextChar = read(state);
    } while (nextChar !== '\n' && nextChar !== null);
    if (nextChar !== null && nextChar !== '\n') {
        unread(state);
    }
    const value = state.input.slice(state.start, state.position + 1);
    return {
        type: 'comment-inline',
        value,
        start: state.start,
        end: state.start + value.length - 1,
    };
}
function scanDollarQuotedString(state) {
    const match = /^(\$[a-zA-Z0-9_]*\$)/.exec(state.input.slice(state.start));
    if (!match) {
        throw new Error('Could not find dollar quoted string opener');
    }
    const label = match[1];
    for (let i = 0; i < label.length - 1; i++) {
        read(state);
    }
    let nextChar = '';
    while (state.input.slice(state.position, state.position + label.length) !== label &&
        nextChar !== null) {
        do {
            nextChar = read(state);
        } while (nextChar !== '$' && nextChar !== null);
        if (nextChar !== '$' && nextChar !== null) {
            unread(state);
        }
    }
    for (let i = 0; i < label.length - 1; i++) {
        read(state);
    }
    const value = state.input.slice(state.start, state.position + 1);
    return {
        type: 'string',
        value,
        start: state.start,
        end: state.start + value.length - 1,
    };
}
function scanString(state, endToken) {
    let nextChar;
    do {
        nextChar = read(state);
        // supporting double quote escaping: 'str''ing'
        if (nextChar === endToken) {
            if (peek(state) === endToken) {
                nextChar = read(state, 1);
            }
        }
    } while (nextChar !== endToken && nextChar !== null);
    if (nextChar !== null && endToken !== nextChar) {
        unread(state);
    }
    const value = state.input.slice(state.start, state.position + 1);
    return {
        type: 'string',
        value,
        start: state.start,
        end: state.start + value.length - 1,
    };
}
function getCustomParam(state, paramTypes) {
    var _a;
    const matches = (_a = paramTypes === null || paramTypes === void 0 ? void 0 : paramTypes.custom) === null || _a === void 0 ? void 0 : _a.map((regex) => {
        const reg = new RegExp(`^(?:${regex})`, 'u');
        return reg.exec(state.input.slice(state.start));
    }).filter((value) => !!value)[0];
    return matches ? matches[0] : null;
}
function scanParameter(state, dialect, paramTypes) {
    var _a, _b, _c;
    const curCh = state.input[state.start];
    const nextChar = peek(state);
    let matched = false;
    if (((_a = paramTypes.numbered) === null || _a === void 0 ? void 0 : _a.length) && paramTypes.numbered.some((type) => type === curCh)) {
        const endIndex = state.input
            .slice(state.start + 1)
            .split('')
            .findIndex((val) => /^\W+/.test(val));
        const maybeNumbers = state.input.slice(state.start + 1, endIndex > 0 ? state.start + endIndex + 1 : state.end + 1);
        if (nextChar !== null && !isNaN(Number(nextChar)) && /^\d+$/.test(maybeNumbers)) {
            let nextChar = null;
            do {
                nextChar = read(state);
            } while (nextChar !== null && !isNaN(Number(nextChar)) && !isWhitespace(nextChar));
            if (nextChar !== null)
                unread(state);
            matched = true;
        }
    }
    if (!matched && ((_b = paramTypes.named) === null || _b === void 0 ? void 0 : _b.length) && paramTypes.named.some((type) => type === curCh)) {
        if (!isQuotedIdentifier(nextChar, dialect)) {
            while (isAlphaNumeric(peek(state)))
                read(state);
            matched = true;
        }
    }
    if (!matched && ((_c = paramTypes.quoted) === null || _c === void 0 ? void 0 : _c.length) && paramTypes.quoted.some((type) => type === curCh)) {
        if (isQuotedIdentifier(nextChar, dialect)) {
            const quoteChar = read(state);
            // end when we reach the end quote
            while ((isAlphaNumeric(peek(state)) || peek(state) === ' ') &&
                peek(state) != ENDTOKENS[quoteChar]) {
                read(state);
            }
            // read the end quote
            read(state);
            matched = true;
        }
    }
    if (!matched && paramTypes.custom && paramTypes.custom.length) {
        const custom = getCustomParam(state, paramTypes);
        if (custom) {
            read(state, custom.length);
            matched = true;
        }
    }
    const value = state.input.slice(state.start, state.position + 1);
    if (!matched && !paramTypes.positional && curCh !== '?') {
        // not positional, panic
        return {
            type: 'unknown',
            value: value,
            start: state.start,
            end: state.start + value.length - 1,
        };
    }
    return {
        type: 'parameter',
        value,
        start: state.start,
        end: state.start + value.length - 1,
    };
}
function scanCommentBlock(state) {
    let nextChar = '';
    let prevChar;
    do {
        prevChar = nextChar;
        nextChar = read(state);
    } while ((prevChar || '') + (nextChar || '') !== '*/' && nextChar !== null);
    if (nextChar !== null && nextChar !== '/') {
        unread(state);
    }
    const value = state.input.slice(state.start, state.position + 1);
    return {
        type: 'comment-block',
        value,
        start: state.start,
        end: state.start + value.length - 1,
    };
}
function scanQuotedIdentifier(state, endToken) {
    let nextChar;
    do {
        nextChar = read(state);
    } while (endToken !== nextChar && nextChar !== null);
    if (nextChar !== null && endToken !== nextChar) {
        unread(state);
    }
    const value = state.input.slice(state.start, state.position + 1);
    return {
        type: 'keyword',
        value,
        start: state.start,
        end: state.start + value.length - 1,
    };
}
function scanWord(state) {
    let nextChar;
    do {
        nextChar = read(state);
    } while (isLetter(nextChar));
    if (nextChar !== null && !isLetter(nextChar)) {
        unread(state);
    }
    const value = state.input.slice(state.start, state.position + 1);
    if (!isKeyword(value)) {
        return skipWord(state, value);
    }
    return {
        type: 'keyword',
        value,
        start: state.start,
        end: state.start + value.length - 1,
    };
}
function scanIndividualCharacter(state) {
    const value = state.input.slice(state.start, state.position + 1);
    const type = resolveIndividualTokenType(value);
    if (!type) {
        return null;
    }
    return {
        type,
        value,
        start: state.start,
        end: state.start + value.length - 1,
    };
}
function skipChar(state) {
    return {
        type: 'unknown',
        value: state.input.slice(state.start, state.position + 1),
        start: state.start,
        end: state.start,
    };
}
function skipWord(state, value) {
    return {
        type: 'unknown',
        value,
        start: state.start,
        end: state.start + value.length - 1,
    };
}
function isWhitespace(ch) {
    return ch === ' ' || ch === '\t' || ch === '\n' || ch === '\r';
}
function isAlphaNumeric(ch) {
    return ch !== null && /\w/.test(ch);
}
function isString(ch, dialect) {
    const stringStart = dialect === 'mysql' ? ["'", '"'] : ["'"];
    return stringStart.includes(ch);
}
function isCustomParam(state, customParamType) {
    return customParamType.some((regex) => {
        const reg = new RegExp(`^(?:${regex})`, 'uy');
        return reg.test(state.input.slice(state.start));
    });
}
function isParameter(ch, state, paramTypes) {
    var _a, _b, _c, _d;
    if (!ch) {
        return false;
    }
    const nextChar = peek(state);
    const prevChar = peekBack(state);
    // HACK (@day): This is a fix for casts and identifiers that use the `::` syntax
    if (ch === ':' && (prevChar === ':' || nextChar === ':')) {
        return false;
    }
    if (paramTypes.positional && ch === '?')
        return true;
    if (((_a = paramTypes.numbered) === null || _a === void 0 ? void 0 : _a.length) && paramTypes.numbered.some((type) => ch === type)) {
        if (nextChar !== null && !isNaN(Number(nextChar))) {
            return true;
        }
    }
    if ((((_b = paramTypes.named) === null || _b === void 0 ? void 0 : _b.length) && paramTypes.named.some((type) => type === ch)) ||
        (((_c = paramTypes.quoted) === null || _c === void 0 ? void 0 : _c.length) && paramTypes.quoted.some((type) => type === ch))) {
        return true;
    }
    if (((_d = paramTypes.custom) === null || _d === void 0 ? void 0 : _d.length) && isCustomParam(state, paramTypes.custom)) {
        return true;
    }
    return false;
}
function isDollarQuotedString(state) {
    return /^\$[\w]*\$/.exec(state.input.slice(state.start)) !== null;
}
function isQuotedIdentifier(ch, dialect) {
    const startQuoteChars = dialect === 'mssql' ? ['"', '['] : ['"', '`'];
    return startQuoteChars.includes(ch);
}
function isCommentInline(ch, state) {
    let isComment = ch === '-';
    if (!isComment) {
        return false;
    }
    // lookahead
    const nextChar = read(state);
    isComment = nextChar === '-';
    if (!isComment) {
        unread(state);
    }
    return isComment;
}
function isCommentBlock(ch, state) {
    let isComment = ch === '/';
    if (!isComment) {
        return false;
    }
    // lookahead
    const nextChar = read(state);
    isComment = nextChar === '*';
    if (!isComment) {
        unread(state);
    }
    return isComment;
}
function isLetter(ch) {
    return ch !== null && ((ch >= 'a' && ch <= 'z') || (ch >= 'A' && ch <= 'Z') || ch === '_');
}
